import logging
import shutil
from collections import defaultdict
# pylint: disable=unused-import
from pathlib import Path

from linky.utils import path_utils
from linky.utils.library_utils import recreate_with_symlinks
from linky.utils.path_utils import get_first_dir, get_paths_in_root, get_path_in_base, get_cat_tag


def tag(paths, cat_tags, config, delete=False, best_effort=False):
    """
    Apply (or delete) multiple tags to multiple paths

    @param paths: The absolute paths. Paths will be wrapped in a string
    @type paths: list[Path] | Path
    @param cat_tags: List of "<category>/<tag>" string
    @type cat_tags: list[basestring] | basestring
    @type config: linky.config.Config
    @param delete: Removes the tag instead of applying it
    @type delete:
    @param best_effort: Ignores errors applying/deleting tags and tries to
                        treat every path
    @type best_effort: bool
    @return:
    @rtype:
    """
    logger = logging.getLogger("tag_multiple")
    if not isinstance(paths, list):
        paths = [paths]
    if not isinstance(cat_tags, list):
        cat_tags = [cat_tags]
    # First clean and separate tags and group them to reduce useless actions
    cleaned_cat_tags = defaultdict(set)
    for cat_tag in cat_tags:
        cat, _tag = [split.strip() for split in cat_tag.split("/")]
        cleaned_cat_tags[cat].add(_tag)

    # Apply tags
    for path in paths:
        # Use the path in the base since tagging might delete the old path
        # in the linked root
        for rel_path_in_base in _iter_path_to_tag(path, config):
            for cat, tags in cleaned_cat_tags.items():
                try:
                    delete = _apply_cat_tags(cat, tags, path, rel_path_in_base, config, delete,
                                             best_effort)
                # pylint: disable=bare-except
                except:
                    if not best_effort:
                        raise
                    logger.warning(
                        "Error ignored while handling path='%s', category='%s', delete='%s'",
                        path, cat, delete
                    )


def _apply_cat_tags(category, tags, path, rel_path_in_base, config, delete, best_effort):
    if category not in config.categories:
        raise ValueError("Unknown category: %s" % category)
    logger = logging.getLogger("tag.apply_cat_tags")
    category_o = config.categories[category]
    other_paths = get_paths_in_root(
        config.base_path / rel_path_in_base, config, category=category
    )
    for _tag in tags:
        try:
            if not category_o.extensible and _tag not in category_o.tags:
                raise ValueError("Unknown tag in %s: %s" % (category, _tag))
            delete, _tag = _check_delete(delete, category_o, _tag, other_paths)
            tag_once(rel_path_in_base, category, _tag, other_paths, config, delete=delete)
        # pylint: disable=bare-except
        except:
            if not best_effort:
                raise
            logger.warning(
                "Error ignored while handling path='%s', category='%s', tag='%s', delete='%s'",
                path, category, _tag, delete
            )
    return delete


def _iter_path_to_tag(path, config):
    """
    Yields files in the given path in order to tag files instead of directories

    @param path:
    @type path: Path
    @param config:
    @type config: linky.config.Config
    @rtype: collections.Iterable[Path]
    """
    if not path.exists():
        raise FileNotFoundError(path)

    base_path = config.base_path
    linked_root = base_path.parent
    current_parent = get_first_dir(path, base_path)
    category_parent = current_parent in config.categories.keys()

    if category_parent:
        distance_from_root = len(path.relative_to(linked_root).parts)

        # Check if we're actually the root
        if distance_from_root == 0:
            raise ValueError("Cannot tag the root")

        # Check whether the given path is a category path
        if distance_from_root == 1:
            raise ValueError("Cannot tag a category")

        # Check whether the given path is a tag path
        if distance_from_root == 2:
            raise ValueError("Cannot tag a tag")
    path_in_base = get_path_in_base(base_path, path, categories=config.categories.keys())
    if path_in_base.is_dir():
        for _abs_path, _ in path_utils.walk(path_in_base, include_dirs=False):
            yield _abs_path.relative_to(base_path)
    else:
        yield path_in_base.relative_to(base_path)


def _check_delete(delete, category_o, tag_str, other_paths):
    if delete:
        # Exclusive tag_once applies default on deletion
        # Removing the last applied tag_once also has the same effect
        if category_o.exclusive or len(other_paths) == 1:
            tag_str = category_o.default
            delete = False
    return delete, tag_str


def tag_once(rel_path_in_base, category_str, tag_str, other_paths, config, delete=False):
    """
    Adds a tag to a file or directory

    :type rel_path_in_base: Path
    :type category_str: basestring
    :type tag_str: basestring
    :type other_paths: list[Path]
    :type config: linky.config.Config
    :param delete: reverses the action to removing the tag_once instead of adding it
    :type delete: bool
    """
    logger = logging.getLogger("tag_once")
    category_o = config.categories[category_str]
    base_path = config.base_path
    linked_root = base_path.parent
    # Build new parent
    new_path = linked_root / category_str / tag_str / rel_path_in_base
    sub_logger = logger.getChild(str(rel_path_in_base))
    if new_path.exists():
        if delete:
            if new_path.is_file():
                new_path.unlink()
            else:
                shutil.rmtree(new_path)
            path_utils.del_empty_parents(new_path)
        else:
            logger.info("Already correctly tagged")
        return

    to_delete = _collect_deletion_candidates(
        base_path, category_o, other_paths, tag_str
    )

    # Actually do the tagging
    recreate_with_symlinks(new_path, base_path / rel_path_in_base)
    sub_logger.info("Tagged %s/%s", category_str, tag_str)

    # Clean up
    if to_delete:
        sub_logger.info("Deleting old tags...")
    for item in to_delete:
        try:
            if item.is_file():
                item.unlink()
            else:
                shutil.rmtree(item, ignore_errors=True)
            path_utils.del_empty_parents(item)
        except FileNotFoundError:
            pass
        if sub_logger.level <= logging.INFO:
            sub_logger.info("\t%s", get_cat_tag(item, base_path.parent))


def _collect_deletion_candidates(base_path, category_o, other_paths, tag_str):
    to_delete = []
    # Handle exclusiveness
    if category_o.exclusive or (tag_str == category_o.default and category_o.exclusive_default):
        # Filter to current category only
        to_delete.extend(other_paths)
    elif category_o.exclusive_default:
        # Delete the exclusive default if it's among the paths
        for other_path in other_paths:
            if get_cat_tag(other_path, base_path.parent).t == category_o.default:
                to_delete.append(other_path)
    return [
        item for item in to_delete
        # Don't delete files that aren't symlinks
        if (item.is_file() and item.is_symlink()) or item.is_dir()
    ]



import unittest
from pathlib import Path

from linky.config import read_conf
from linky.utils import path_utils
from linky.utils.path_utils import (
    BASE_NAME, get_dir_prefix, get_first_dir, get_path_in_base,
    get_paths_in_root, get_prefixed_path, is_path_in_base,
)
from tests import TEST_DATA_PATH


# pylint: disable=missing-function-docstring
class GetPathInBaseTest(unittest.TestCase):
    """
    Make sure path_utils.get_path_in_base is working well
    """

    def setUp(self) -> None:
        self.base_path = Path("/home/lol/videos/.base")

    def _test_bad_path(self, target_path, error_message):
        with self.assertRaises(ValueError) as ass:
            get_path_in_base(self.base_path, Path(target_path))
        self.assertEqual(ass.exception.args[0], error_message)

    def _test_good_path(self, target_path, expected_result, categories=None, relative=False):
        self.assertEqual(
            get_path_in_base(
                self.base_path, Path(target_path),
                categories=categories, relative=relative),
            Path(expected_result)
        )

    def test_path_normal(self):
        self._test_good_path(
            "/home/lol/videos/comedian/wonderboy/jokes/on/you/test.avi",
            "/home/lol/videos/.base/jokes/on/you/test.avi",
            categories=["comedian"]
        )

    def test_path_normal_relative(self):
        self._test_good_path(
            "/home/lol/videos/comedian/wonderboy/jokes/on/you/test.avi",
            "jokes/on/you/test.avi",
            categories=["comedian"],
            relative=True
        )

    def test_path_root(self):
        self._test_bad_path(
            "/home/lol/videos/",
            "Cannot target the link root"
        )

    def test_path_no_common_path(self):
        self._test_bad_path(
            "/home/lol/music/nothing",
            "Cannot target path outside of root"
        )

    def test_path_within_base(self):
        path_in_base = "/home/lol/videos/.base/somewhere/else/in/base"
        self._test_good_path(path_in_base, path_in_base)

    def test_path_within_base_relative(self):
        self._test_good_path(
            "/home/lol/videos/.base/somewhere/else/in/base",
            "somewhere/else/in/base", relative=True
        )

    def test_path_is_base(self):
        self._test_bad_path(
            self.base_path,
            "Cannot target the base path"
        )

    def test_path_with_relative_links(self):
        self._test_bad_path(
            "./videos/watched/somewhere/else/bah.mkv",
            "Paths have to be absolute"
        )


class GetDirPrefixTest(unittest.TestCase):
    """
    Tests for path_utils.get_dir_prefix
    """

    def test_long_name(self):
        self.assertListEqual(
            get_dir_prefix("Avadakedavra"),
            ["A", "Ava"]
        )

    def test_short_names(self):
        for name in ["Lol", "Lo"]:
            self.assertListEqual(
                get_dir_prefix(name),
                ["L"]
            )

    def test_no_prefix_names(self):
        for name in ["B", ""]:
            self.assertListEqual(
                get_dir_prefix(name),
                []
            )


class GetPrefixedTest(unittest.TestCase):
    """
    Tests for path_utils.get_prefixed_path
    """

    def test_simple(self):
        self.assertEqual(
            get_prefixed_path(Path("/mnt/Videos/Gazing Stars")),
            Path("/mnt/Videos/G/Gaz/Gazing Stars")
        )


class FindBaseTest(unittest.TestCase):
    """
    Tests for path_utils.find_base
    """

    def test_find_base(self):
        expected_base = TEST_DATA_PATH / "simple_base" / path_utils.BASE_NAME
        base = path_utils.find_base(TEST_DATA_PATH / "simple_base" / "dir1" / "dir2")
        self.assertEqual(base, expected_base)

    def test_find_from_file(self):
        expected_base = TEST_DATA_PATH / "simple_base" / path_utils.BASE_NAME
        base = path_utils.find_base(TEST_DATA_PATH / "simple_base" / "dir1" / ".gitkeep")
        self.assertEqual(base, expected_base)

    def test_find_base_max_iteration(self):
        self.assertRaises(
            StopIteration,
            path_utils.find_base, TEST_DATA_PATH / "simple_base" / "dir1" / "dir2", 1
        )

    def test_find_base_no_base(self):
        self.assertRaises(
            (ValueError, FileNotFoundError),
            path_utils.find_base, Path("/usr/lib/dpkg/methods")
        )


class FirstDirTest(unittest.TestCase):
    """
    Tests for path_utils.get_first_dir
    """

    def test_category(self):
        first_dir = path_utils.get_first_dir(TEST_DATA_PATH / "simple_base" / "dir1" / "dir2")
        self.assertEqual(first_dir, "dir1")


class GetPathsInRoot(unittest.TestCase):
    """
    Tests for path_utils.get_paths_in_root
    """

    def setUp(self) -> None:
        self.test_dir = TEST_DATA_PATH / "movies"
        self.base_path = self.test_dir / BASE_NAME
        self.config = read_conf(self.test_dir)

    def test_valid_path_in_base(self):
        path = self.base_path / "The Hurt Locker (2008)" / "The.Hurt.Locker.2008.1080p.BRrip.mkv"

        self._test_valid_path(path, 5)

    def _test_valid_path(self, path, expected_count, category=None):
        paths = get_paths_in_root(path, self.config, category)
        path_count = len(paths)
        self.assertEqual(path_count, expected_count)
        for path_in_root in paths:
            self.assertFalse(
                is_path_in_base(path_in_root, self.base_path),
                "Returned path is in base: %s" % path_in_root
            )

    def test_valid_path_outside_base(self):
        path = self.test_dir / \
               "Watched" / "Watched" / \
               "The Hurt Locker (2008)" / "The.Hurt.Locker.2008.1080p.BRrip.mkv"
        self._test_valid_path(path, 5)

    def test_should_pass_with_same_category_limited_to_category(self):
        path = self.test_dir / \
               "Actors" / "Jeremy Renner" / \
               "The Hurt Locker (2008)" / "The.Hurt.Locker.2008.1080p.BRrip.mkv"
        self._test_valid_path(path, 3, "Actors")

    def test_should_pass_with_different_category_limited_to_category(self):
        path = self.test_dir / \
               "Watched" / "Watched" / \
               "The Hurt Locker (2008)" / "The.Hurt.Locker.2008.1080p.BRrip.mkv"
        self._test_valid_path(path, 3, "Actors")

    def test_with_base(self):
        path = self.test_dir / path_utils.BASE_NAME
        self.assertRaisesRegex(
            ValueError, "Can't target base",
            get_paths_in_root, path, self.config, self.base_path
        )

    def test_with_link_root(self):
        path = self.test_dir
        self.assertRaisesRegex(
            ValueError, "Can't target linked root",
            get_paths_in_root, path, self.config, self.base_path
        )

    def test_with_tag(self):
        self.assertRaisesRegex(
            ValueError, "Path cannot be tag group or tag",
            get_paths_in_root, self.test_dir / "Actors", self.config
        )

    def test_with_tag_group(self):
        self.assertRaisesRegex(
            ValueError, "Path cannot be tag group or tag",
            get_paths_in_root, self.test_dir / "Actors" / "Emilia Clarke", self.config
        )

    def test_invalid_path_in_base(self):
        path = self.base_path / "I don't exist"

        self.assertRaisesRegex(
            ValueError, "Path doesn't exist",
            get_paths_in_root, path, self.config, self.base_path
        )


class GetFirstDir(unittest.TestCase):
    """
    Tests for path_utils.get_first_dir
    """

    def test_should_pass_with_base(self):
        base_path = Path("/mnt/storage/") / BASE_NAME
        self.assertEqual(
            get_first_dir(base_path, base_path),
            BASE_NAME
        )

    def test_should_fail_with_root(self):
        base_path = Path("/mnt/storage/.base")
        self.assertRaisesRegex(
            ValueError, "Can't get first directory from root",
            get_first_dir, base_path.parent, base_path,
        )


if __name__ == "__main__":
    unittest.main()



import shutil

from linky.utils import path_utils


class DupeHandler:
    """
    A base class for handling duplicate paths
    """
    NAME = None

    def __call__(self, path, new_path):
        raise NotImplementedError()


class KeepDupeHandler(DupeHandler):
    """
    Keeps the destination and removes the source.
    """

    NAME = "keep"

    def __call__(self, path, new_path):
        """
        @type path: pathlib.Path
        @type new_path: pathlib.Path
        """
        if path.is_file():
            path.unlink()
        else:
            shutil.rmtree(path)


class MergeDupeHandler(DupeHandler):
    """
    Merges directories.
    Files will be moved and empty directories will be ignored
    """
    NAME = "merge"

    def __call__(self, path, new_path):
        """
        @type path: pathlib.Path
        @type new_path: pathlib.Path
        """
        if not path.is_dir():
            raise ValueError("Source is not a directory", path)
        if not new_path.is_dir():
            raise ValueError("Destination is not a directory", new_path)

        for abs_item, rel_item in path_utils.walk(path):
            abs_new_item = new_path / rel_item

            # Ignore directories and older source files
            if abs_item.is_dir() or \
                    (abs_new_item.is_file() and
                     abs_new_item.stat().st_mtime < abs_item.stat().st_mtime):
                continue
            # Move files
            abs_item.parent.mkdir(parents=True, exist_ok=True)
            shutil.move(abs_item, abs_new_item)
        shutil.rmtree(path)


class ReplaceDupeHandler(DupeHandler):
    """
    Replaces the destination with a source
    """
    NAME = "replace"

    def __call__(self, path, new_path):
        """
        @type new_path: pathlib.Path
        """
        shutil.move(str(path), str(new_path))


class ReplaceOldDupeHandler(DupeHandler):
    """
    Replaces the destination if the source is new
    """
    NAME = "replace_old"

    def __call__(self, path, new_path):
        """
        @type new_path: pathlib.Path
        """
        if new_path.stat().st_mtime > path.stat().st_mtime:
            return
        shutil.move(str(path), str(new_path))


DUPE_HANDLERS = {
    c.NAME: c
    for c in (
        KeepDupeHandler(),
        MergeDupeHandler(),
        ReplaceDupeHandler(),
        ReplaceOldDupeHandler(),
    )}



import argparse

from cliff.command import Command

from linky.actions import move
from linky.utils.path_utils import abs_path


class MinArgsAction(argparse.Action):
    """
    A validator for options that need a minimum amount of values.
    """
    def __init__(self, option_strings, dest, **kwargs):
        self.min_count = int(kwargs.pop("nargs"))
        kwargs["nargs"] = "+"
        super().__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        if len(values) < self.min_count:
            raise argparse.ArgumentError(self, "At least %s required" % self.min_count)
        setattr(namespace, self.dest, values)


class MoveCommand(Command):
    """
    Move a file or directory.
    The change will be mirrored in the base.
    """

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "paths", nargs=2, action=MinArgsAction, type=abs_path,
            help="At least 2 paths.\n"
                 "More than 2 paths means move these SOURCES into DIR"
        )
        return parser

    def take_action(self, parsed_args):
        move(*parsed_args.paths)



import logging

from cliff.command import Command

from linky.actions import add
from linky.config import read_conf
from linky.utils import path_utils
from linky.utils.dupe_handlers import DUPE_HANDLERS


class AddCommand(Command):
    """
    Add an item to the linky management
    """

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument("-b", "--base-path",
                            type=path_utils.abs_path,
                            help="Where to add the file/dir."
                                 "Can also be a path with a parent containing a base path."
                            )
        parser.add_argument("-d", "--dupe-handler",
                            choices=DUPE_HANDLERS.keys(),
                            help="How an existing destination / dupe is handled\n" +
                            "\n".join(
                                " - %s: %s" % (name, c.__doc__.strip())
                                for name, c in sorted(DUPE_HANDLERS.items())
                            ))
        group = parser.add_mutually_exclusive_group()
        group.add_argument("-l", "--linked-root-prefix",
                           action="store_true",
                           help="Calculates the prefix in the linked root"
                                "Category and tag will be calculated."
                                "e.g"
                                "  path = linked_root/categoryA/tagA/dir/subdir/filename"
                                "  -->prefix = dir/subdir"
                                "  -->result = base_path/dir/subdir/filename"
                                "  -->category = categoryA"
                                "  -->tag = tagA"
                           )
        group.add_argument("-p", "--prefix",
                           default="",
                           help="<base_path>/<prefix>/<path name>\n"
                                "e.g\n"
                                "  path = /tmp/dir/subdir/filename\n"
                                "  prefix = just/a/prefix\n"
                                "  result --> base_path/just/a/prefix/filename\n")
        parser.add_argument("paths",
                            nargs="+", type=path_utils.abs_path,
                            help="The base path will be guessed from these paths "
                                 "if `-b` isn't provided!")
        return parser

    def take_action(self, parsed_args):
        logger = logging.getLogger("commands.add")
        for _abs_path in parsed_args.paths:
            try:
                base_path = path_utils.find_base(parsed_args.base_path or _abs_path)
                add(
                    _abs_path,
                    base_path,
                    read_conf(base_path),
                    prefix=parsed_args.prefix,
                    linked_root_prefix=parsed_args.linked_root_prefix
                )
            # pylint:disable=broad-except
            except Exception as exception:
                logger.warning("Couldn't add '%s': %s", _abs_path, exception)



import logging
import shutil

from linky.actions.tag import tag
from linky.utils.dupe_handlers import DUPE_HANDLERS, DupeHandler, MergeDupeHandler, \
    ReplaceDupeHandler
from linky.utils.library_utils import move_to_base, recreate_with_symlinks
from linky.utils.path_utils import get_path_in_base, get_paths_in_root, get_cat_tag


def add(path, base_path, config, cat_tag=None, prefix="", linked_root_prefix=False,
        dupe_handler_name=None):
    """
    Imports a path into the base to be managed by linky

    @param path: What we're trying to import
    @type path: Path
    @param base_path: The base into which the item should be imported
    @type base_path: Path
    @type config: linky.config.Config
    @param cat_tag: The predefined tag for a specific category
    @type cat_tag: linky.utils.path_utils.CategoryTagTuple
    @param prefix: An additional prefix to add to the path
                   If config.prefix_at_import is used in addition to this,
                   this prefix will first be added and the standard prefix
                   calculated thereafter
    @type prefix: basestring
    @param linked_root_prefix: Mutually exclusive with prefix and cat_tag.
                                Calculates the prefix within the linked root
                                as well as the category and tag
    @type linked_root_prefix: bool
    @param dupe_handler_name: How dupes will be handled
    @type dupe_handler_name: DupeHandler
    """
    logger = logging.getLogger("actions.add.add")
    logger.info("Adding: %s", path)
    logger.debug("Base path: %s", base_path)

    if linked_root_prefix:
        prefix = get_path_in_base(
            base_path, path,
            categories=config.categories.keys(), relative=True
        ).parent
        cat_tag = get_cat_tag(path, config.base_path.parent)
        logger.debug("Adding linked root prefix '%s'", prefix)

    if cat_tag:
        logger.info("Pre cat-tag: %s", cat_tag)

    dupe_handler = DUPE_HANDLERS.get(dupe_handler_name)  # type: DupeHandler
    try:
        path_in_base = move_to_base(path, base_path, config, prefix)
    except ValueError as value_error:
        if not dupe_handler:
            raise
        path_in_base = value_error.args[1]
        logger.info("Handling dupe with method %s", dupe_handler.NAME)
        dupe_handler(path, path_in_base)
    logger.debug("New path in base '%s'", path_in_base)
    if config.categories:
        if isinstance(dupe_handler, (MergeDupeHandler, ReplaceDupeHandler)):
            logger.debug("Modified existing item in base. Deleting existing tags...")
            _remove_other_paths(path_in_base, config, logger)

        logger.debug("Categorizing and tagging...")
        # Apply default tags for everything
        # except when cat_tag is passed
        for cat_name, category in config.categories.items():
            tag_name = category.default
            if cat_tag is not None and cat_name == cat_tag.c:
                tag_name = cat_tag.t
            tag(path_in_base, "%s/%s" % (cat_name, tag_name), config)
    else:
        logger.info("Simply relinking in root...")
        recreate_with_symlinks(base_path.parent / path_in_base.relative_to(base_path), path_in_base)


def _remove_other_paths(path_in_base, config, logger):
    for other_path in get_paths_in_root(path_in_base, config):
        # Don't delete files
        if not (other_path.is_symlink() or other_path.exists()):
            continue
        if other_path.is_file():
            other_path.unlink()
        else:
            shutil.rmtree(other_path)
        logger.debug("Deleted %s", other_path)



import os
import re
from collections import namedtuple
from os.path import abspath, commonpath
from pathlib import Path

BASE_NAME = ".base"
CONFIG_DIR_NAME = ".linky"

RESERVED_DIR_NAMES = (BASE_NAME, CONFIG_DIR_NAME)

ALNUM_REGEX = re.compile(r"[a-z0-9]", re.IGNORECASE)


class CategoryTagTuple(namedtuple("_CategoryTagTuple", ["c", "t"])):
    """
    A specialized tuple to hold a category and a tag.

    It simply has a specialized __str__ function
    """

    def __str__(self):
        return "%s/%s" % (self.c, self.t)


def get_dir_prefix(name):
    """
    Calculates the 1 to 2 directory prefix for a name

    Example:

        Avadakedavra --> [ A , Ava]
        Lol --> [L]
        Ba --> [B]
        '' --> []

    :param name: A file or folder name
    :type name: basestring
    :return: prefix for long enough names or empty list
    :rtype: list[str]
    """
    res = "".join(ALNUM_REGEX.findall(name))
    prefix = []
    res_len = len(res)
    if res_len > 1:
        comp_1 = res[0].upper()
        if res_len <= 3:
            prefix = [comp_1]
        else:
            prefix = [comp_1, res[:3].capitalize()]

    return prefix


def get_prefixed_path(path):
    """
    Prefix a path name with a linky directory prefix

    >>>> get_prefixed_path(Path("/opt/link_root/some_item")
    >>>> # Path("/opt/link_root/S/Som/some_item")

    @see: get_dir_prefix
    @type path: Path
    @rtype: Path
    """
    return Path(*[
        path.parent,
        *get_dir_prefix(path.name),
        path.name
    ])


def get_path_in_base(base_path, target, categories=None, relative=False):
    """

    :param base_path:
    :type base_path: Path
    :param target:
    :type target: Path
    :type categories: list | tuple | None
    :param relative: Make sure the path we return is relative to the base
    :type relative: bool
    :rtype: Path
    """
    if not (base_path.is_absolute() and target.is_absolute()):
        raise ValueError("Paths have to be absolute")

    if target == base_path:
        raise ValueError("Cannot target the base path")
    if target == base_path.parent:
        raise ValueError("Cannot target the link root")
    if str(target).startswith(str(base_path / CONFIG_DIR_NAME)):
        raise ValueError("Cannot target the config dir or its contents")

    common_path = Path(commonpath([base_path, target]))
    if common_path == base_path:
        if relative:
            return target.relative_to(base_path)
        return target
    try:
        relative_target = target.relative_to(base_path.parent)
    except ValueError as origal_error:
        raise ValueError("Cannot target path outside of root") from origal_error

    cut_depth = 0
    if is_path_in_base(target, base_path):
        # Being in
        cut_depth = 1
    elif categories is not None and relative_target.parts[0] in categories:
        # Is the target manage or not
        # Being in a category means, it has to be
        # So we ignore the first 2 directories: category and tag
        cut_depth = 2

    ret = Path(*[
        base_path if not relative else "",
        *relative_target.parts[cut_depth:]
    ])
    return ret


def is_path_in_base(path, base_path=None, check_exist=True):
    """
    :type path: Path
    :type base_path: Path | None
    :param check_exist: Additional check whether the path actually exists
    :type check_exist: bool
    """
    if check_exist and not path.exists():
        return False

    if base_path:
        try:
            path.relative_to(base_path)
            return True
        # pylint: disable=bare-except
        except:
            return False
    else:
        for part in reversed(path.parts):
            if part == BASE_NAME:
                return True
        return False


def get_paths_in_root(path, config, category=None):
    """
    Finds other symlinks to the given file in the linked root.

    @param path: When given a path outside of the base, the corresponding path in base is used
    @type path: Path
    @type config: Config
    @param category: Limits the search of other symlinks to this category
    @type category: basestring | None
    @rtype: list[Path]
    """
    base_path = _check_path_and_category(path, category, config)

    link_root = base_path.parent
    category_names = config.categories.keys()
    paths = []

    if is_path_in_base(path, base_path):
        rel_path = path.relative_to(base_path)
    else:
        first_parent = get_first_dir(path, base_path)
        rel_path = path.relative_to(link_root)
        # Is it tagged?
        if first_parent in category_names:
            # Shouldn't be a tag group or tag!
            if len(rel_path.parts) < 3:
                raise ValueError("Path cannot be tag group or tag", path)
            # Remove category and tag
            rel_path = Path(*rel_path.parts[2:])

    def add_possible_path(root):
        possible_path = root / rel_path
        if possible_path.exists():
            paths.append(possible_path)

    for item in link_root.iterdir():
        if item.name == BASE_NAME:
            continue
        # Filter by category if requested
        if category and item.name != category:
            continue
        if item.name in category_names:
            for tag in item.iterdir():
                add_possible_path(tag)
        else:
            add_possible_path(item)
    return paths


def _check_path_and_category(path, category, config):
    if not path.exists():
        raise ValueError("Path doesn't exist", path)

    if not path.is_absolute():
        raise ValueError("Path must be absolute", path)

    base_path = config.base_path
    if path == base_path:
        raise ValueError("Can't target base", path)
    if path == base_path.parent:
        raise ValueError("Can't target linked root", path)
    if category is not None and category not in config.categories:
        raise ValueError("Unknown category")
    return base_path


def find_base(current_dir, max_count=None, current_count=1):
    """
    Find first parent with a basedir

    :type current_dir: Path
    :type max_count: int
    :type current_count: int
    :rtype: Path
    """
    if current_dir.is_file():
        return find_base(current_dir.parent, max_count, current_count)
    if max_count and current_count > max_count:
        raise StopIteration("Maximum height reached without finding base")
    for entry in current_dir.iterdir():
        if entry.is_dir() and entry.name == BASE_NAME:
            return entry
    if str(current_dir) == current_dir.root:
        raise ValueError("Reached root without finding base")
    return find_base(current_dir.parent, max_count, current_count + 1)


def get_first_dir(path, base_path=None):
    """

    :param path:
    :type path: Path
    :param base_path:
    :type base_path:
    :return:
    :rtype: basestring
    """
    if not path.is_absolute():
        raise ValueError("Please provide an absolute path")
    if not base_path:
        base_path = find_base(path)
    root = base_path.parent
    if path == root:
        raise ValueError("Can't get first directory from root")
    rel_path = path.relative_to(root)
    first = rel_path.parts[0]

    return first


def get_path_props(path):
    """
    Compile select information about a path

    @type path: Path
    @rtype: dict
    """
    return {
        "is_symlink": path.is_symlink(),
        "is_dir": path.is_dir(),
        "is_file": path.is_file(),
    }


def get_cat_tag(path, link_root):
    """
    Guess the category and tag of a path in the link root

    We assume it actually is a categorized/tagged path in the linked root

    @type path: Path
    @type link_root: Path
    @rtype: CategoryTagTuple
    """
    rel = path.relative_to(link_root)
    return CategoryTagTuple(rel.parts[0], rel.parts[1])


def iter_linked_root(root_path, config):
    """
    Iterates over a linked root only returning managed (or to be managed) elements.

    That means it will return the contents of the directory and category-tag directories.
    It will NOT return:

     - category directories themselves
     - tag directories themselves
     - the base directory itself
     - nor the base directory contents

    @type root_path: Path
    @type config: Config
    """
    category_dict = config.categories
    for entry in root_path.iterdir():
        if entry.name == BASE_NAME:
            continue
        # Entry category folder
        if entry.is_dir() and entry.name in category_dict:
            for tag_dir in entry.iterdir():
                for tagged_entry in tag_dir.iterdir():
                    yield tagged_entry, CategoryTagTuple(entry.name, tag_dir.name)
        else:
            yield entry, None


def walk(path, topdown=True, include_dirs=True, include_files=True):
    """
    Simplified os.walk yielding the absolute and relative path per item

    @type path: Path
    @param topdown: Whether to navigate from the top of the directory structure
    @type topdown: bool
    @param include_dirs: Adds directories to the items to yield in each directory
    @type include_dirs: bool
    @param include_files: Adds files to the items to yield in each directory
    @type include_files: bool
    """
    for current_dir, dirnames, filenames in os.walk(path, topdown=topdown):
        current_dir = Path(current_dir)
        rel_root = current_dir.relative_to(path)
        items = []
        if include_dirs:
            items += dirnames
        if include_files:
            items += filenames
        for item in items:
            yield current_dir / item, rel_root / item


def abs_path(path):
    """
    @type path: basestring
    @rtype: Path
    """
    return Path(abspath(path))


def del_empty_parents(path):
    """
    Deletes the parent of the given path if it's empty,
    then tries to delete the parent's parent, etc.

    @type path: Path
    """
    try:
        path.parent.rmdir()
        del_empty_parents(path.parent)
    # pylint:disable=bare-except
    except:
        pass


def normpath(path):
    """
    Normalizes a path using os.path.normpath and returns a Path

    os.path.normpath returns a string

    @type path: Path
    @rtype: Path
    """
    return Path(os.path.normpath(path))



"""
linky
Copyright (C) 2021 LoveIsGrief

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from pathlib import Path
from typing import List

from linky.config import Config
from linky.utils.path_utils import get_paths_in_root


def get_tags(config: Config, *paths: Path, relative: bool = False) -> List[Path]:
    """
    Get all tags of a the given paths

    @param config: The repository's config
    @param paths: These must be within the linked root, otherwise an exception will be thrown
    @param relative: Whether the paths should be printed relative to the linked root
    @return: string (not natural!) sorted list of absolute or linked-root relative paths
    """
    tag_paths = []
    for path in paths:
        tag_paths.extend(in_root for in_root in get_paths_in_root(path, config))
    if relative:
        tag_paths = [tag_path.relative_to(config.base_path.parent) for tag_path in tag_paths]
    return sorted(tag_paths, key=str)



import logging
import shutil
from pathlib import Path

from linky.actions import add
from linky.config import Config, read_conf
from linky.utils import path_utils
from linky.utils.path_utils import BASE_NAME, iter_linked_root, CONFIG_DIR_NAME


def init(root_path, overwrite_config=False):
    """
    Creates the base directory and moves everything else inside

    If there are already category dirs then the move will be followed by symlinking
     the contents of those category folders.

    :param root_path:
    :type root_path: Path
    :param overwrite_config: Whether a config in the link root
                              should overwrite the one in the base
    :type overwrite_config: bool
    """
    logger = logging.getLogger("init")
    if not root_path.is_dir():
        raise ValueError("Given path must be a directory")

    base_dir = root_path / BASE_NAME
    if base_dir.is_dir():
        raise ValueError("Given path has already been initialized")

    logger.info("Initializing...")
    base_dir.mkdir(parents=True, exist_ok=True)
    logger.info("Base dir: %s", base_dir)

    # Try and read a config first (we need categories)
    config = _get_config_for_init(root_path, base_dir, overwrite_config)
    failed_files = []
    for root_entry, cat_tag in iter_linked_root(root_path, config):
        # Add file by file
        if root_entry.is_file():
            _add(root_entry, base_dir, config, cat_tag, failed_files)
        else:
            for abs_item, rel_item in path_utils.walk(root_entry, topdown=False):
                if not abs_item.is_file():
                    # Clean empty dirs we encounter
                    _clean_if_empty(abs_item)
                    continue
                _add(
                    abs_item, base_dir, config, cat_tag, failed_files,
                    prefix=Path(root_entry.name) / rel_item.parent
                )

                # Clean up empty parents
                _clean_if_empty(abs_item.parent)
    if failed_files:
        for failed_file, exception in failed_files:
            logger.warning("Failed to import '%s': %s", failed_file, exception)
        logger.warning("You can try importing the failed files with `linky add $filename` "
                       "to get more information about the error")
    logger.info("Done...")


def _add(abs_item, base_dir, config, cat_tag, failed_files, prefix=None):
    try:
        add(abs_item, base_dir, config, cat_tag, prefix=prefix)
    # pylint: disable=broad-except
    except Exception as ex:
        failed_files.append((abs_item, ex))


def _clean_if_empty(path, ignore_errors=True):
    if path.is_dir() and len(list(path.iterdir())) == 0:
        shutil.rmtree(path, ignore_errors=ignore_errors)


def _get_config_for_init(link_root, base_dir, overwrite_config):
    """
    Tries to read the config from the link root or base path.

    The one in the link will override an existing config in the base path.
    But that's only possible by passing @param overwrite_config


    @type link_root: Path
    @type base_dir: Path
    @type overwrite_config: bool
    @return: An initialized config if one was detected
    @rtype: Config
    """
    possible_config_dir = link_root / CONFIG_DIR_NAME
    config_dir = base_dir / CONFIG_DIR_NAME
    config = Config(base_dir)

    if possible_config_dir.is_dir():
        # Replace config if possible
        if config_dir.exists():
            if overwrite_config:
                shutil.rmtree(config_dir)
            else:
                raise ValueError("New configuration found. "
                                 "Pass --overwrite-config to supersede old config")
        shutil.move(possible_config_dir, config_dir)

    if config_dir.exists():
        config = read_conf(link_root)
    return config



import logging

from importlib import resources
import yamale
import yaml
from yamale.schema import Schema

from linky.utils.path_utils import find_base, CONFIG_DIR_NAME


class Config:
    """
    Container for the configuration of a link root
    """
    FILE_CATEGORIES = "categories.yaml"
    FILE_MAIN = "main.yaml"

    def __init__(self, base_path):
        self._log = logging.getLogger("Config")
        self._categories = {}
        self.prefix_at_import = False
        self.base_path = base_path

    @property
    def categories(self):
        """
        :returns dict[basestring,Category]
        """
        return self._categories

    @categories.setter
    def categories(self, value):
        for name, props in value.items():
            self._categories[name] = Category(**props)
        self._log.debug("Known categories: %s", ",".join(self._categories.keys()))


class Category:
    """
    A container for a category's configuration
    """

    def __init__(self,
                 default=None,
                 extensible=False,
                 exclusive_default=False,
                 exclusive=False,
                 tags=None
                 ):
        self.default = default
        self.extensible = extensible
        self.exclusive_default = exclusive_default
        self.exclusive = exclusive
        self.tags = tags or []


def _get_schema(name):
    schema_stream = resources.files("linky").joinpath("schemas/%s.schema.yaml" % name)
    try:
        loader = yaml.CSafeLoader
    except AttributeError:  # System does not have libyaml
        loader = yaml.SafeLoader

    with schema_stream.open("rb") as stream:
        raw_schemas = list(yaml.load_all(stream, Loader=loader))
        # First document is the base schema
        try:
            schema = Schema(raw_schemas[0], name)
            # Additional documents contain Includes.
            for raw_schema in raw_schemas[1:]:
                schema.add_include(raw_schema)
        except (TypeError, SyntaxError) as ex:
            error = "Schema error in file %s\n" % name
            error += str(ex)
            raise SyntaxError(error) from ex
        return schema


def read_conf(path):
    """
    :type path: Path
    :returns Config
    """
    base_path = find_base(path)
    config_dir_path = base_path / CONFIG_DIR_NAME

    config = Config(base_path)
    configs_d = {}
    # Validate the configs
    for filename in (config.FILE_MAIN, config.FILE_CATEGORIES):
        config_path = config_dir_path / filename
        if not config_path.is_file():
            continue
        schema = _get_schema(filename.split(".", maxsplit=1)[0])
        data = yamale.make_data(config_path)
        yamale.validate(schema, data)
        configs_d[filename] = data[0][0]

    main_config = configs_d.get(config.FILE_MAIN)
    if main_config:
        config.prefix_at_import = main_config.get("prefix_at_import", False)

    categories_config = configs_d.get(config.FILE_CATEGORIES)
    if categories_config:
        config.categories = categories_config

    return config



import re


def camel2kebab(string):
    """
    Transforms CamelCase to kebab-case

    Adapted from http://stackoverflow.com/questions/1796180/ddg#1796247

    :type string: basestring
    :rtype: basestring
    """
    _string = re.sub("(.)([A-Z][a-z]+)", r"\1-\2", string)
    return re.sub("([a-z0-9])([A-Z])", r"\1-\2", _string).lower()



import logging
import shutil
from pathlib import Path

from linky.config import read_conf
from linky.utils import path_utils


def remove(path: Path, _all: bool = False):
    """
    Permanently removes either files or directories from linky management

    @param path: What to delete
    @param _all: Only impacts directories: whether to select items that have the current tag
                  or all items regardless of tag.
                 This is essentially the difference between walking through the current directory
                  or walking through the directory in the base.
    """
    if _all:
        remove_all(path)
    else:
        remove_files(path)


def remove_all(path: Path):
    """
    Permanently removes a file or directory from linky management

    @param path: What we're trying to delete
    """
    logger = logging.getLogger("actions.remove")
    logger.debug("Removing: %s", path)
    config = read_conf(path)
    base_path = config.base_path
    logger.debug("Base path: %s", base_path)

    path_in_base = path_utils.get_path_in_base(base_path, path, config.categories.keys())
    other_paths = path_utils.get_paths_in_root(path_in_base, config)

    for other_path in other_paths + [path_in_base]:
        if other_path.is_symlink() or other_path.is_file():
            other_path.unlink()
        elif other_path.is_dir():
            shutil.rmtree(other_path)
        logger.info("Removed %s", other_path)

        # Make sure to remove empty parent directories
        path_utils.del_empty_parents(other_path)


def remove_files(path: Path):
    """
    Only removes files within the path from linky management

    Folders will be recursed into to remove the files they contain.
    Using this approach, it isn't possible to delete folders across tags or tag groups.

    For example
    .
    └── Tags
        ├── Document
        │   └── folder
        │       ├── document.odt
        │       └── useful_document.odt
        ├── Movie
        │   └── folder
        │       └── movie.mp4
        ├── Music
        │   └── folder
        │       └── music.m4a
        └── Useful
            └── folder
                ├── useful
                └── useful_document.odt
    Targeting Tags/Document/folder will only remove:
      - Tags/Document/folder/useful_document.odt
      - Tags/Document/folder/document.odt
      - Tags/Document/folder
      - Tags/Useful/folder/useful_document.odt (it's a file and has other tags)
    Nothing else.

    @param path: What we're trying to delete
    """
    logger = logging.getLogger("actions.remove_files")
    logger.debug("Removing file/files in: %s", path)
    config = read_conf(path)
    base_path = config.base_path
    logger.debug("Base path: %s", base_path)

    path_in_base = path_utils.get_path_in_base(base_path, path, config.categories.keys())

    if path_in_base == path:
        # Deleting it in base is the equivalent of wanting to remove it everywhere
        remove_all(path)

    elif path.is_file():
        # Files can be deleted everywhere
        other_paths = path_utils.get_paths_in_root(path_in_base, config)

        for other_path in other_paths + [path_in_base]:
            other_path.unlink()
            logger.info("Removed %s", other_path)
            # Make sure to remove empty parent directories
            path_utils.del_empty_parents(other_path)

    elif path.is_dir():
        for abs_child, _ in path_utils.walk(path, include_dirs=False):
            remove_files(abs_child)

    else:
        raise ValueError("Path is neither file nor directory")



import errno
import os
import shutil
from os.path import relpath
from pathlib import Path

from linky.utils.path_utils import walk, get_dir_prefix


def move_to_base(path, base_path, config, prefix=""):
    """
    Move an item into the base

    @param path: what to move
    @type path: Path
    @type base_path: Path
    @type config: linky.config.Config
    @param prefix: What to prefix the item with in the base
    @type prefix: basestring
    @return: The new path in the base
    @rtype: Path
    """
    if not path.exists():
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), str(path))
    rel_path = Path(path.name)
    if prefix:
        rel_path = Path(prefix) / rel_path
    if config.prefix_at_import:
        rel_path = Path(*[
            *get_dir_prefix(rel_path.parts[0]),
            rel_path
        ])
    new_path = base_path / rel_path

    new_path.parent.mkdir(parents=True, exist_ok=True)
    if new_path.exists():
        raise ValueError("Destination exists. Please choose a dupe handler", new_path)
    shutil.move(path, new_path)

    return new_path


def recreate_with_symlinks(path, symlink_target):
    """
    Recreates a file or directory structure in a new location.
    All files will be relative symlinks

    @param path: The new location
    @type path: Path
    @param symlink_target: What to recreate
    @type symlink_target: Path
    """
    if symlink_target.is_dir():
        for abs_base_item, rel_base_item in walk(symlink_target):
            new_path = (path / rel_base_item)
            if abs_base_item.is_dir():
                # Recreate directory
                new_path.mkdir(parents=True, exist_ok=True)
            else:
                # Link to new path in base
                create_relative_symlink(new_path, abs_base_item)
    elif symlink_target.is_file():
        create_relative_symlink(path, symlink_target)
    else:
        raise ValueError("Symlink target doesn't exist", symlink_target)


def create_relative_symlink(path, symlink_target):
    """
    Creates a relative symlink at the given path.

    @type path: Path
    @type symlink_target: Path | basestring
    """
    path.parent.mkdir(parents=True, exist_ok=True)
    link_path = relpath(symlink_target, path.parent)
    if path.exists():
        if path.resolve(False) == symlink_target:
            return
        path.unlink()
    path.symlink_to(link_path)



from cliff.command import Command

from linky.actions.tags import get_tags
from linky.config import read_conf
from linky.utils.path_utils import abs_path


class TagsCommand(Command):
    """
    Get tags of a given file
    """

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        parser.add_argument("-a", "--absolute",
                            action="store_true",
                            help="Print absolute paths")
        parser.add_argument("paths", nargs="+", type=abs_path)

        return parser

    def take_action(self, parsed_args):
        config = read_conf(parsed_args.paths[0])
        tag_paths = get_tags(config, *parsed_args.paths, relative=not parsed_args.absolute)
        print("\n".join(sorted(str(tag_path) for tag_path in tag_paths)))



from cliff.command import Command

from linky.actions import init
from linky.utils.path_utils import abs_path


class InitCommand(Command):
    """
    Initializes a directory it it hasn't been done before
    """

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument("--overwrite-config",
                            action="store_true",
                            help="Overwrites existing configuration if "
                                 "a new configuration directory is found in "
                                 "the provided path")
        parser.add_argument("path")
        return parser

    def take_action(self, parsed_args):
        init(abs_path(parsed_args.path), parsed_args.overwrite_config)



from cliff.command import Command

from linky.actions import tag
from linky.config import read_conf
from linky.utils.path_utils import abs_path


class TagCommand(Command):
    """
    Tags a given file
    """

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        parser.add_argument("-d", "--delete",
                            action="store_true",
                            help="Deletes the tags from the file")
        parser.add_argument("-b", "--best-effort",
                            action="store_true",
                            help="Ignores errors and tries to apply changes to each given path")
        parser.add_argument("-t", "--tags",
                            required=True,
                            action="append",
                            help="The category and tag to use separated by a '/' e.g Rating/1\n"
                                 "Can be repeated as many times as necessary")
        parser.add_argument("paths", nargs="+", type=abs_path)

        return parser

    def take_action(self, parsed_args):
        config = read_conf(parsed_args.paths[0])
        tag(
            parsed_args.paths,
            parsed_args.tags,
            config,
            parsed_args.delete,
            parsed_args.best_effort,
        )



import logging
import shutil
from pathlib import Path

from linky.config import read_conf
from linky.utils.library_utils import recreate_with_symlinks
from linky.utils.path_utils import (
    del_empty_parents, find_base, get_cat_tag, get_first_dir,
    get_path_in_base, get_paths_in_root, is_path_in_base, normpath,
)

logger = logging.getLogger("move")


def move(*paths):
    """

    @param paths: At least 2 paths.
                  >2 means the last path will be interpreted as a directory
    @type paths: list[Path]
    @return:
    @rtype:
    """
    if len(paths) < 2:
        raise ValueError("Need at least 2 paths")
    sources = paths[:-1]
    new_path = paths[-1]

    if len(sources) == 1:
        _move(sources[0], new_path)
    else:
        for source in sources:
            _move(source, new_path, target_is_dir=True)


def _move(path, new_path, target_is_dir=False):
    base_path = find_base(path)

    config = read_conf(path)
    categories = config.categories.keys()

    # When moving into a directory tack on the name of item being moved
    target_is_dir = target_is_dir or new_path.is_dir()
    if target_is_dir:
        new_path = new_path / path.name

    _check_move_params(path, new_path, base_path, categories)

    old_rel_path = get_path_in_base(base_path, path, categories, True)
    try:
        new_rel_path = normpath(get_path_in_base(base_path, new_path, categories, True))
    except ValueError as original_error:
        raise ValueError("Bad move target", new_path) from original_error

    path_in_base = base_path / old_rel_path
    other_links = get_paths_in_root(path, config)

    # First move the file or directory in the base
    new_path_in_base = base_path / new_rel_path
    new_path_in_base.parent.mkdir(parents=True, exist_ok=True)
    logger.info("Moving from '%s' to '%s'", old_rel_path, new_rel_path)
    logger.debug("Moving from '%s' to '%s'", path_in_base, new_path_in_base)
    shutil.move(str(path_in_base), str(new_path_in_base))

    del_empty_parents(path_in_base)

    # Then move the other links in the link root
    for other_link in other_links:
        # Remove the relative path from the other link
        # that gets the tag dir or the link root
        dir_root = Path(
            *other_link.parts[:-len(old_rel_path.parts)]
        )
        new_other_path = dir_root / new_rel_path

        logger.info("Moving in other link: '%s' -> '%s'", other_link, new_other_path)
        recreate_with_symlinks(new_other_path, new_path_in_base)

        # Cleanup old path
        if new_path_in_base.is_file():
            other_link.unlink()
        elif new_path_in_base.is_dir():
            shutil.rmtree(other_link)
        del_empty_parents(other_link)


def _check_move_params(path, new_path, base_path, categories):
    """
    Make sure an attempt is being made to make an unsupported move:

     into or out of the root, base, category or tag
    @type path: Path
    @type new_path: Path
    @type base_path: Path
    @type categories: list[basestring]
    """
    in_base = is_path_in_base(path, base_path)
    new_in_base = is_path_in_base(new_path, base_path, False)
    if in_base != new_in_base:
        raise ValueError("Trying to move in or out of base")
    if in_base:
        return
    old_parent = get_first_dir(path, base_path)
    new_parent = get_first_dir(new_path, base_path)

    old_categorized = old_parent in categories
    new_categorized = new_parent in categories
    if old_categorized != new_categorized:
        raise ValueError("Cannot categorize using move")
    if old_categorized and new_categorized:
        old_tag = get_cat_tag(path, base_path.parent)
        new_tag = get_cat_tag(new_path, base_path.parent)
        if old_tag != new_tag:
            raise ValueError("Cannot switch categories or tag using move")



import setuptools

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()
with open("requirements.txt", "r") as fh:
    REQUIREMENTS = fh.readlines()
with open("linky/VERSION", "r") as fh:
    VERSION = fh.read().strip()

setuptools.setup(
    name="linky-db",
    version=VERSION,
    author="LoveIsGrief",
    author_email="loveisgrief@tuta.io",
    description="A script to manage and tag files using symlinks",
    install_requires=REQUIREMENTS,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/NamingThingsIsHard/linky",
    packages=setuptools.find_packages(
        exclude=["tests", "tests.*"]
    ),
    package_data={
        "linky": [
            "VERSION",
            "schemas/*.yaml"
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Operating System :: POSIX :: Linux",
    ],
    project_urls={
        "Bug Tracker": "https://gitlab.com/NamingThingsIsHard/linky/issues",
    },
    entry_points={
        "console_scripts": [
            "linky = linky.app:main",
        ]
    }
)



import inspect
import logging
import sys
from importlib import import_module

from importlib import resources
from cliff.app import App
from cliff.commandmanager import CommandManager

from linky.utils.str_utils import camel2kebab

COMMAND_SUFFIX = "Command"
VERSION = resources.files(__package__).joinpath("VERSION").read_bytes().decode().strip()


class LinkyCommandManager(CommandManager):
    """
    Customized manager to load linky commands
    """

    def load_commands(self, namespace):
        """
        Creates commands from all *Command classes in the namespace

        :type namespace: basestring
        """
        module = import_module(namespace)
        for name, o in inspect.getmembers(module):
            if not (inspect.isclass(o) and name.endswith(COMMAND_SUFFIX)):
                continue

            # We make sure to remove the redundant "Command" suffix
            self.add_command(
                camel2kebab(name[:-len(COMMAND_SUFFIX)]),
                o
            )


App.NAME = "linky"


# pylint: disable=missing-class-docstring
class LinkyApp(App):
    CONSOLE_MESSAGE_FORMAT = logging.BASIC_FORMAT

    def __init__(self):
        super().__init__(
            description="Helps keep files organized by linking them to a base folder",
            version=VERSION,
            command_manager=LinkyCommandManager("linky.commands"),
        )


# pylint: disable=missing-function-docstring
def main(argv=None):
    argv = argv or sys.argv[1:]
    myapp = LinkyApp()
    if len(argv) == 0:
        argv = ["help"]
    return myapp.run(argv)


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



import logging

from cliff.command import Command

from linky.actions import remove
from linky.utils import path_utils


class RemoveCommand(Command):
    """
    Permanently removes / deletes a file or folder and all its links.
    Warning: This is a NON-REVERSIBLE action!

    Notice:
    When deleting a directory in a tag folder, this will ONLY remove / delete the
    files and directories WITH the tag. Those without the tag will not be touched.
    To delete an entire folder regardless of tag, use -a/--all
    """

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument("-a", "--all", action="store_true",
                            help="When passed a directory, delete the ENTIRE directory, "
                                 "regardless of tags.")
        parser.add_argument("paths", nargs="+", type=path_utils.abs_path,
                            help="The file or directory to remove")
        return parser

    def take_action(self, parsed_args):
        _log = logging.getLogger("command.remove")
        should_remove_all = parsed_args.all
        for _abs_path in parsed_args.paths:
            if _abs_path.exists():
                try:
                    remove(_abs_path, should_remove_all)
                # pylint:disable=broad-except
                except Exception as exception:
                    _log.error("Couldn't remove '%s': %s",
                               _abs_path, exception,
                               exc_info=True)
            else:
                _log.info("Ignored nonexistent: %s", _abs_path)



from linky.commands.add import AddCommand
from linky.commands.init import InitCommand
from linky.commands.move import MoveCommand
from linky.commands.remove import RemoveCommand
from linky.commands.tag import TagCommand
from linky.commands.tags import TagsCommand


__all__ = [
    "AddCommand",
    "InitCommand",
    "MoveCommand",
    "RemoveCommand",
    "TagCommand",
    "TagsCommand",
]
